package net.knarcraft.armoredelytra.update;

import com.google.common.base.Preconditions;
import com.google.gson.JsonElement;
import com.google.gson.JsonObject;
import com.google.gson.JsonParser;
import com.google.gson.JsonSyntaxException;
import net.knarcraft.armoredelytra.ArmoredElytra;
import org.bukkit.plugin.java.JavaPlugin;

import java.io.IOException;
import java.io.InputStreamReader;
import java.net.HttpURLConnection;
import java.net.URL;
import java.time.Instant;
import java.util.concurrent.CompletableFuture;
import java.util.logging.Level;

/**
 * A utility class to assist in checking for updates for plugins uploaded to
 * <a href="https://spigotmc.org/resources/">SpigotMC</a>. Before any members of
 * this class are accessed, {@link #init(ArmoredElytra, int)} must be invoked by the plugin, preferably in its {@link
 * JavaPlugin#onEnable()} method, though that is not a requirement.
 * <p>
 * This class performs asynchronous queries to
 * <a href="https://spiget.org">SpiGet</a>, an REST server which is updated
 * periodically. If the results of {@link #requestUpdateCheck()} are inconsistent with what is published on SpigotMC, it
 * may be due to SpiGet's cache. Results will be updated in due time.
 * <p>
 * Some modifications were made to support downloading of updates and storing the age of an update.
 *
 * @author Parker Hawke - 2008Choco
 */
public final class UpdateChecker {

    private static final String USER_AGENT = "ArmoredElytra-update-checker";
    private static final String UPDATE_URL = "https://api.spiget.org/v2/resources/%d/versions?size=1&sort=-releaseDate";

    private static UpdateChecker instance;

    UpdateResult lastResult = null;

    final ArmoredElytra plugin;
    private final int pluginID;
    private final VersionScheme versionScheme;

    private UpdateChecker(final ArmoredElytra plugin, final int pluginID, final VersionScheme versionScheme) {
        this.plugin = plugin;
        this.pluginID = pluginID;
        this.versionScheme = versionScheme;
    }

    /**
     * Requests an update check to SpiGet. This request is asynchronous and may not complete immediately as an HTTP GET
     * request is published to the SpiGet API.
     *
     * @return a future update result
     */
    public CompletableFuture<UpdateResult> requestUpdateCheck() {
        return CompletableFuture.supplyAsync(
                () -> {
                    int responseCode;
                    try {
                        URL url = new URL(String.format(UPDATE_URL, pluginID));
                        HttpURLConnection connection = (HttpURLConnection) url.openConnection();
                        connection.addRequestProperty("User-Agent", USER_AGENT);

                        InputStreamReader reader = new InputStreamReader(connection.getInputStream());
                        responseCode = connection.getResponseCode();

                        JsonElement element = new JsonParser().parse(reader);
                        if (!element.isJsonArray()) {
                            return new UpdateResult(this, UpdateReason.INVALID_JSON);
                        }

                        reader.close();

                        JsonObject versionObject = element.getAsJsonArray().get(0).getAsJsonObject();

                        long age = -1;
                        String ageString = versionObject.get("releaseDate").getAsString();
                        try {
                            age = getAge(Long.parseLong(ageString));
                        } catch (NumberFormatException e) {
                            plugin.myLogger(Level.WARNING,
                                    "Failed to obtain age of update from ageString: \"" + ageString + "\"");
                        }

                        String current = plugin.getDescription().getVersion(), newest = versionObject.get("name")
                                .getAsString();
                        String latest = versionScheme.compareVersions(current, newest);

                        if (latest == null) {
                            return new UpdateResult(this, UpdateReason.UNSUPPORTED_VERSION_SCHEME);
                        } else if (latest.equals(current)) {
                            return new UpdateResult(this, current.equals(newest) ?
                                    UpdateReason.UP_TO_DATE :
                                    UpdateReason.UNRELEASED_VERSION, current, age);
                        } else if (latest.equals(newest)) {
                            return new UpdateResult(this, UpdateReason.NEW_UPDATE, latest, age);
                        }
                    } catch (IOException e) {
                        return new UpdateResult(this, UpdateReason.COULD_NOT_CONNECT);
                    } catch (JsonSyntaxException e) {
                        return new UpdateResult(this, UpdateReason.INVALID_JSON);
                    }

                    return new UpdateResult(this, responseCode == 401 ?
                            UpdateReason.UNAUTHORIZED_QUERY : UpdateReason.UNKNOWN_ERROR);
                });
    }

    /**
     * Gets the difference in seconds between a given time and the current time.
     *
     * @param updateTime A moment in time to compare the current time to.
     * @return The difference in seconds between a given time and the current time.
     */
    private long getAge(final long updateTime) {
        long currentTime = Instant.now().getEpochSecond();
        return currentTime - updateTime;
    }

    /**
     * Gets the last update result that was queried by {@link #requestUpdateCheck()}. If no update check was performed
     * since this class' initialization, this method will return null.
     *
     * @return the last update check result. null if none.
     */
    public UpdateResult getLastResult() {
        return lastResult;
    }

    /**
     * Initializes this update checker with the specified values and return its instance. If an instance of
     * UpdateChecker has already been initialized, this method will act similarly to {@link #get()} (which is
     * recommended after initialization).
     *
     * @param plugin        the plugin for which to check updates. Cannot be null
     * @param pluginID      <p>The ID of the plugin as identified in the SpigotMC resource link. For example,
     *                      "<a href="https://www.spigotmc.org/resources/veinminer/12038/">
     *                      https://www.spigotmc.org/resources/veinminer/<b>12038</b>/</a>" would expect "12038" as a
     *                      value. The value must be greater than 0</p>
     * @param versionScheme a custom version scheme parser. Cannot be null
     * @return the UpdateChecker instance
     */
    public static UpdateChecker init(final ArmoredElytra plugin, final int pluginID, final VersionScheme versionScheme) {
        Preconditions.checkArgument(pluginID > 0, "Plugin ID must be greater than 0");

        return (instance == null) ? instance = new UpdateChecker(plugin, pluginID, versionScheme) : instance;
    }

    /**
     * Initializes this update checker with the specified values and return its instance. If an instance of
     * UpdateChecker has already been initialized, this method will act similarly to {@link #get()} (which is
     * recommended after initialization).
     *
     * @param plugin   the plugin for which to check updates. Cannot be null
     * @param pluginID <p>The ID of the plugin as identified in the SpigotMC resource link. For example,
     *                 "<a href="https://www.spigotmc.org/resources/veinminer/12038/">
     *                 https://www.spigotmc.org/resources/veinminer/<b>12038</b>/</a>" would expect "12038" as a value. The
     *                 value must be greater than 0</p>
     * @return the UpdateChecker instance
     */
    public static UpdateChecker init(final ArmoredElytra plugin, final int pluginID) {
        return init(plugin, pluginID, new VersionScheme());
    }

    /**
     * Gets the initialized instance of UpdateChecker. If {@link #init(ArmoredElytra, int)} has not yet been invoked,
     * this method will throw an exception.
     *
     * @return the UpdateChecker instance
     */
    public static UpdateChecker get() {
        Preconditions.checkState(instance != null,
                "Instance has not yet been initialized. Be sure #init() has been invoked");
        return instance;
    }

    /**
     * Checks whether the UpdateChecker has been initialized or not (if {@link #init(ArmoredElytra, int)} has been
     * invoked) and {@link #get()} is safe to use.
     *
     * @return true if initialized, false otherwise
     */
    public static boolean isInitialized() {
        return instance != null;
    }

}
